;; Port of http://www.cl.cam.ac.uk/~mgk25/ucs/wcwidth.c
(library (yuni text width)
         (export
           string-width
           string-width/cjk
           char-width
           char-width/cjk)
         (import (rnrs)
                 (rnrs r5rs))

(define (width proc s)
  (fold-left + 0 (map (lambda (c) (proc c))
                      (string->list s))))
(define (string-width s)
  (width char-width s))
(define (string-width/cjk s)
  (width char-width/cjk s))

;; Util
(define (bisearch code tbl) ;; bool
  (define max (- (vector-length tbl) 1))
  (define (val x) (vector-ref tbl x))
  (define (itr min max)
    (define mid (quotient (+ min max) 2))
    (cond
      ((> min max) #f)
      ((> code (cdr (val mid)))
       (itr (+ mid 1) max))
      ((< code (car (val mid)))
       (itr min (- mid 1)))
      (else #t)))
  (and (not (or (< code (car (val 0)))
                (> code (cdr (val max)))))
       (itr 0 max)))

(define (char-width c) ;; => 1, #f, 0, 2
  (define code (char->integer c))
  (define (zone x y) (and (>= code x) (<= code y)))
  (cond
    ((= code 0) ;; Null char
     0)
    ((or (< code 32) (and (< code #xa0) (>= code #x7f)))
     #f)
    ((bisearch code combining) 0)
    ((and (>= code #x1100)
          (or 
            (<= code #x115f) ;; Hangul Jamo init.
            (= code #x2329)
            (= code #x232a)
            (and (>= code #x2e80)
                 (<= code #xa4cf)
                 (not (= code #x303f)))
            (zone #xac00 #xd7a3) ;; Hangul Syllables
            (zone #xf900 #xfaff) ;; CJK Compat. Ideographs
            (zone #xfe10 #xfe19) ;; Vertial forms
            (zone #xfe30 #xfe6f) ;; CJK Compatibility forms
            (zone #xff00 #xff60) ;; Fullwidth forms
            (zone #xffe0 #xffe6) 
            (zone #x20000 #x2fffd)
            (zone #x30000 #x3fffd)))
     2)
    (else 1)))

(define (char-width/cjk c)
  (define code (char->integer c))
  (if (bisearch code ambiguous)
    2
    (char-width c)))

;; Data
(define combining
  #( 
     ( #x0300 . #x036F )  ( #x0483 . #x0486 )  ( #x0488 . #x0489 ) 
     ( #x0591 . #x05BD )  ( #x05BF . #x05BF )  ( #x05C1 . #x05C2 ) 
     ( #x05C4 . #x05C5 )  ( #x05C7 . #x05C7 )  ( #x0600 . #x0603 ) 
     ( #x0610 . #x0615 )  ( #x064B . #x065E )  ( #x0670 . #x0670 ) 
     ( #x06D6 . #x06E4 )  ( #x06E7 . #x06E8 )  ( #x06EA . #x06ED ) 
     ( #x070F . #x070F )  ( #x0711 . #x0711 )  ( #x0730 . #x074A ) 
     ( #x07A6 . #x07B0 )  ( #x07EB . #x07F3 )  ( #x0901 . #x0902 ) 
     ( #x093C . #x093C )  ( #x0941 . #x0948 )  ( #x094D . #x094D ) 
     ( #x0951 . #x0954 )  ( #x0962 . #x0963 )  ( #x0981 . #x0981 ) 
     ( #x09BC . #x09BC )  ( #x09C1 . #x09C4 )  ( #x09CD . #x09CD ) 
     ( #x09E2 . #x09E3 )  ( #x0A01 . #x0A02 )  ( #x0A3C . #x0A3C ) 
     ( #x0A41 . #x0A42 )  ( #x0A47 . #x0A48 )  ( #x0A4B . #x0A4D ) 
     ( #x0A70 . #x0A71 )  ( #x0A81 . #x0A82 )  ( #x0ABC . #x0ABC ) 
     ( #x0AC1 . #x0AC5 )  ( #x0AC7 . #x0AC8 )  ( #x0ACD . #x0ACD ) 
     ( #x0AE2 . #x0AE3 )  ( #x0B01 . #x0B01 )  ( #x0B3C . #x0B3C ) 
     ( #x0B3F . #x0B3F )  ( #x0B41 . #x0B43 )  ( #x0B4D . #x0B4D ) 
     ( #x0B56 . #x0B56 )  ( #x0B82 . #x0B82 )  ( #x0BC0 . #x0BC0 ) 
     ( #x0BCD . #x0BCD )  ( #x0C3E . #x0C40 )  ( #x0C46 . #x0C48 ) 
     ( #x0C4A . #x0C4D )  ( #x0C55 . #x0C56 )  ( #x0CBC . #x0CBC ) 
     ( #x0CBF . #x0CBF )  ( #x0CC6 . #x0CC6 )  ( #x0CCC . #x0CCD ) 
     ( #x0CE2 . #x0CE3 )  ( #x0D41 . #x0D43 )  ( #x0D4D . #x0D4D ) 
     ( #x0DCA . #x0DCA )  ( #x0DD2 . #x0DD4 )  ( #x0DD6 . #x0DD6 ) 
     ( #x0E31 . #x0E31 )  ( #x0E34 . #x0E3A )  ( #x0E47 . #x0E4E ) 
     ( #x0EB1 . #x0EB1 )  ( #x0EB4 . #x0EB9 )  ( #x0EBB . #x0EBC ) 
     ( #x0EC8 . #x0ECD )  ( #x0F18 . #x0F19 )  ( #x0F35 . #x0F35 ) 
     ( #x0F37 . #x0F37 )  ( #x0F39 . #x0F39 )  ( #x0F71 . #x0F7E ) 
     ( #x0F80 . #x0F84 )  ( #x0F86 . #x0F87 )  ( #x0F90 . #x0F97 ) 
     ( #x0F99 . #x0FBC )  ( #x0FC6 . #x0FC6 )  ( #x102D . #x1030 ) 
     ( #x1032 . #x1032 )  ( #x1036 . #x1037 )  ( #x1039 . #x1039 ) 
     ( #x1058 . #x1059 )  ( #x1160 . #x11FF )  ( #x135F . #x135F ) 
     ( #x1712 . #x1714 )  ( #x1732 . #x1734 )  ( #x1752 . #x1753 ) 
     ( #x1772 . #x1773 )  ( #x17B4 . #x17B5 )  ( #x17B7 . #x17BD ) 
     ( #x17C6 . #x17C6 )  ( #x17C9 . #x17D3 )  ( #x17DD . #x17DD ) 
     ( #x180B . #x180D )  ( #x18A9 . #x18A9 )  ( #x1920 . #x1922 ) 
     ( #x1927 . #x1928 )  ( #x1932 . #x1932 )  ( #x1939 . #x193B ) 
     ( #x1A17 . #x1A18 )  ( #x1B00 . #x1B03 )  ( #x1B34 . #x1B34 ) 
     ( #x1B36 . #x1B3A )  ( #x1B3C . #x1B3C )  ( #x1B42 . #x1B42 ) 
     ( #x1B6B . #x1B73 )  ( #x1DC0 . #x1DCA )  ( #x1DFE . #x1DFF ) 
     ( #x200B . #x200F )  ( #x202A . #x202E )  ( #x2060 . #x2063 ) 
     ( #x206A . #x206F )  ( #x20D0 . #x20EF )  ( #x302A . #x302F ) 
     ( #x3099 . #x309A )  ( #xA806 . #xA806 )  ( #xA80B . #xA80B ) 
     ( #xA825 . #xA826 )  ( #xFB1E . #xFB1E )  ( #xFE00 . #xFE0F ) 
     ( #xFE20 . #xFE23 )  ( #xFEFF . #xFEFF )  ( #xFFF9 . #xFFFB ) 
     ( #x10A01 . #x10A03 )  ( #x10A05 . #x10A06 )  ( #x10A0C . #x10A0F ) 
     ( #x10A38 . #x10A3A )  ( #x10A3F . #x10A3F )  ( #x1D167 . #x1D169 ) 
     ( #x1D173 . #x1D182 )  ( #x1D185 . #x1D18B )  ( #x1D1AA . #x1D1AD ) 
     ( #x1D242 . #x1D244 )  ( #xE0001 . #xE0001 )  ( #xE0020 . #xE007F ) 
     ( #xE0100 . #xE01EF )))
(define ambiguous 
  #(
    ( #x00A1 . #x00A1 )  ( #x00A4 . #x00A4 )  ( #x00A7 . #x00A8 ) 
    ( #x00AA . #x00AA )  ( #x00AE . #x00AE )  ( #x00B0 . #x00B4 ) 
    ( #x00B6 . #x00BA )  ( #x00BC . #x00BF )  ( #x00C6 . #x00C6 ) 
    ( #x00D0 . #x00D0 )  ( #x00D7 . #x00D8 )  ( #x00DE . #x00E1 ) 
    ( #x00E6 . #x00E6 )  ( #x00E8 . #x00EA )  ( #x00EC . #x00ED ) 
    ( #x00F0 . #x00F0 )  ( #x00F2 . #x00F3 )  ( #x00F7 . #x00FA ) 
    ( #x00FC . #x00FC )  ( #x00FE . #x00FE )  ( #x0101 . #x0101 ) 
    ( #x0111 . #x0111 )  ( #x0113 . #x0113 )  ( #x011B . #x011B ) 
    ( #x0126 . #x0127 )  ( #x012B . #x012B )  ( #x0131 . #x0133 ) 
    ( #x0138 . #x0138 )  ( #x013F . #x0142 )  ( #x0144 . #x0144 ) 
    ( #x0148 . #x014B )  ( #x014D . #x014D )  ( #x0152 . #x0153 ) 
    ( #x0166 . #x0167 )  ( #x016B . #x016B )  ( #x01CE . #x01CE ) 
    ( #x01D0 . #x01D0 )  ( #x01D2 . #x01D2 )  ( #x01D4 . #x01D4 ) 
    ( #x01D6 . #x01D6 )  ( #x01D8 . #x01D8 )  ( #x01DA . #x01DA ) 
    ( #x01DC . #x01DC )  ( #x0251 . #x0251 )  ( #x0261 . #x0261 ) 
    ( #x02C4 . #x02C4 )  ( #x02C7 . #x02C7 )  ( #x02C9 . #x02CB ) 
    ( #x02CD . #x02CD )  ( #x02D0 . #x02D0 )  ( #x02D8 . #x02DB ) 
    ( #x02DD . #x02DD )  ( #x02DF . #x02DF )  ( #x0391 . #x03A1 ) 
    ( #x03A3 . #x03A9 )  ( #x03B1 . #x03C1 )  ( #x03C3 . #x03C9 ) 
    ( #x0401 . #x0401 )  ( #x0410 . #x044F )  ( #x0451 . #x0451 ) 
    ( #x2010 . #x2010 )  ( #x2013 . #x2016 )  ( #x2018 . #x2019 ) 
    ( #x201C . #x201D )  ( #x2020 . #x2022 )  ( #x2024 . #x2027 ) 
    ( #x2030 . #x2030 )  ( #x2032 . #x2033 )  ( #x2035 . #x2035 ) 
    ( #x203B . #x203B )  ( #x203E . #x203E )  ( #x2074 . #x2074 ) 
    ( #x207F . #x207F )  ( #x2081 . #x2084 )  ( #x20AC . #x20AC ) 
    ( #x2103 . #x2103 )  ( #x2105 . #x2105 )  ( #x2109 . #x2109 ) 
    ( #x2113 . #x2113 )  ( #x2116 . #x2116 )  ( #x2121 . #x2122 ) 
    ( #x2126 . #x2126 )  ( #x212B . #x212B )  ( #x2153 . #x2154 ) 
    ( #x215B . #x215E )  ( #x2160 . #x216B )  ( #x2170 . #x2179 ) 
    ( #x2190 . #x2199 )  ( #x21B8 . #x21B9 )  ( #x21D2 . #x21D2 ) 
    ( #x21D4 . #x21D4 )  ( #x21E7 . #x21E7 )  ( #x2200 . #x2200 ) 
    ( #x2202 . #x2203 )  ( #x2207 . #x2208 )  ( #x220B . #x220B ) 
    ( #x220F . #x220F )  ( #x2211 . #x2211 )  ( #x2215 . #x2215 ) 
    ( #x221A . #x221A )  ( #x221D . #x2220 )  ( #x2223 . #x2223 ) 
    ( #x2225 . #x2225 )  ( #x2227 . #x222C )  ( #x222E . #x222E ) 
    ( #x2234 . #x2237 )  ( #x223C . #x223D )  ( #x2248 . #x2248 ) 
    ( #x224C . #x224C )  ( #x2252 . #x2252 )  ( #x2260 . #x2261 ) 
    ( #x2264 . #x2267 )  ( #x226A . #x226B )  ( #x226E . #x226F ) 
    ( #x2282 . #x2283 )  ( #x2286 . #x2287 )  ( #x2295 . #x2295 ) 
    ( #x2299 . #x2299 )  ( #x22A5 . #x22A5 )  ( #x22BF . #x22BF ) 
    ( #x2312 . #x2312 )  ( #x2460 . #x24E9 )  ( #x24EB . #x254B ) 
    ( #x2550 . #x2573 )  ( #x2580 . #x258F )  ( #x2592 . #x2595 ) 
    ( #x25A0 . #x25A1 )  ( #x25A3 . #x25A9 )  ( #x25B2 . #x25B3 ) 
    ( #x25B6 . #x25B7 )  ( #x25BC . #x25BD )  ( #x25C0 . #x25C1 ) 
    ( #x25C6 . #x25C8 )  ( #x25CB . #x25CB )  ( #x25CE . #x25D1 ) 
    ( #x25E2 . #x25E5 )  ( #x25EF . #x25EF )  ( #x2605 . #x2606 ) 
    ( #x2609 . #x2609 )  ( #x260E . #x260F )  ( #x2614 . #x2615 ) 
    ( #x261C . #x261C )  ( #x261E . #x261E )  ( #x2640 . #x2640 ) 
    ( #x2642 . #x2642 )  ( #x2660 . #x2661 )  ( #x2663 . #x2665 ) 
    ( #x2667 . #x266A )  ( #x266C . #x266D )  ( #x266F . #x266F ) 
    ( #x273D . #x273D )  ( #x2776 . #x277F )  ( #xE000 . #xF8FF ) 
    ( #xFFFD . #xFFFD )  ( #xF0000 . #xFFFFD )  ( #x100000 . #x10FFFD )
    ))

)
